import asyncio
from collections import defaultdict
from collections.abc import Callable, Iterator
from dataclasses import dataclass, field

from fastcs.attribute_io_ref import AttributeIORef
from fastcs.attributes import ONCE, Attribute, AttrR
from fastcs.cs_methods import Command, Scan, ScanCallback
from fastcs.logging import logger as _fastcs_logger
from fastcs.tracer import Tracer

tracer = Tracer(name=__name__)
logger = _fastcs_logger.bind(logger_name=__name__)


@dataclass
class ControllerAPI:
    """Attributes, bound methods and sub APIs of a `Controller`"""

    path: list[str] = field(default_factory=list)
    """Path within controller tree (empty if this is the root)"""
    attributes: dict[str, Attribute] = field(default_factory=dict)
    command_methods: dict[str, Command] = field(default_factory=dict)
    scan_methods: dict[str, Scan] = field(default_factory=dict)
    sub_apis: dict[str, "ControllerAPI"] = field(default_factory=dict)
    """APIs of the sub controllers of the `Controller` this API was built from"""
    description: str | None = None

    def walk_api(self) -> Iterator["ControllerAPI"]:
        """Walk through all the nested `ControllerAPI` s of this `ControllerAPI`.

        Yields the `ControllerAPI` s from a depth-first traversal of the tree,
        including self.

        """
        yield self
        for api in self.sub_apis.values():
            yield from api.walk_api()

    def __repr__(self):
        return f"""\
ControllerAPI(path={self.path}, sub_apis=[{", ".join(self.sub_apis.keys())}])\
"""

    def get_scan_and_initial_coros(
        self,
    ) -> tuple[list[ScanCallback], list[ScanCallback]]:
        scan_dict: dict[float, list[Callable]] = defaultdict(list)
        initial_coros: list[Callable] = []

        for controller_api in self.walk_api():
            _add_scan_method_tasks(scan_dict, initial_coros, controller_api)
            _add_attribute_update_tasks(scan_dict, initial_coros, controller_api)

        scan_coros = _get_periodic_scan_coros(scan_dict)
        return scan_coros, initial_coros


def _add_scan_method_tasks(
    scan_dict: dict[float, list[Callable]],
    initial_coros: list[Callable],
    controller_api: ControllerAPI,
):
    for method in controller_api.scan_methods.values():
        if method.period is ONCE:
            initial_coros.append(method.fn)
        else:
            scan_dict[method.period].append(method.fn)


def _add_attribute_update_tasks(
    scan_dict: dict[float, list[Callable]],
    initial_coros: list[Callable],
    controller_api: ControllerAPI,
):
    for attribute in controller_api.attributes.values():
        match attribute:
            case (
                AttrR(_io_ref=AttributeIORef(update_period=update_period)) as attribute
            ):
                if update_period is ONCE:
                    initial_coros.append(attribute.bind_update_callback())
                elif update_period is not None:
                    scan_dict[update_period].append(attribute.bind_update_callback())


def _get_periodic_scan_coros(
    scan_dict: dict[float, list[Scan]],
) -> list[ScanCallback]:
    periodic_scan_coros: list[ScanCallback] = []
    for period, methods in scan_dict.items():
        periodic_scan_coros.append(_create_periodic_scan_coro(period, methods))

    return periodic_scan_coros


def _create_periodic_scan_coro(period: float, scans: list[Scan]) -> ScanCallback:
    async def _sleep():
        await asyncio.sleep(period)

    methods = [_sleep] + scans  # Create periodic behavior

    async def scan_coro() -> None:
        while True:
            await asyncio.gather(*[method() for method in methods])

    return scan_coro



import asyncio

from fastcs import FastCS
from fastcs.attributes import AttrR, AttrW
from fastcs.controller import Controller
from fastcs.datatypes import Bool, Int
from fastcs.transport.epics.ca.transport import EpicsCATransport
from fastcs.transport.epics.options import EpicsIOCOptions
from fastcs.transport.rest.options import RestServerOptions
from fastcs.transport.rest.transport import RestTransport
from fastcs.transport.tango.options import TangoDSROptions
from fastcs.transport.tango.transport import TangoTransport


class MyTestController(Controller):
    read_int: AttrR = AttrR(Int(), initial_value=0)
    write_bool: AttrW = AttrW(Bool())


def run():
    transport_options = [
        RestTransport(rest=RestServerOptions(port=8090)),
        EpicsCATransport(
            ca_ioc=EpicsIOCOptions(pv_prefix="BENCHMARK-DEVICE"),
        ),
        TangoTransport(dsr=TangoDSROptions(dev_name="MY/BENCHMARK/DEVICE")),
    ]
    instance = FastCS(MyTestController(), transport_options, asyncio.get_event_loop())
    instance.run()


if __name__ == "__main__":
    run()



import pytest

from fastcs.attributes import AttrR
from fastcs.controller import Controller
from fastcs.datatypes import Float, Int


def test_controller_nesting():
    controller = Controller()
    sub_controller = Controller()
    sub_sub_controller = Controller()

    controller.a = sub_controller
    sub_controller.b = sub_sub_controller

    assert sub_controller.path == ["a"]
    assert sub_sub_controller.path == ["a", "b"]
    assert controller.sub_controllers == {"a": sub_controller}
    assert sub_controller.sub_controllers == {"b": sub_sub_controller}

    with pytest.raises(ValueError, match=r"existing sub controller"):
        controller.a = Controller()

    with pytest.raises(ValueError, match=r"already registered"):
        controller.c = sub_controller


class SomeSubController(Controller):
    def __init__(self):
        super().__init__()

    sub_attribute = AttrR(Int())

    root_attribute = AttrR(Int())


class SomeController(Controller):
    annotated_attr_not_defined_in_init: AttrR[int]
    equal_attr = AttrR(Int())
    annotated_and_equal_attr: AttrR[int] = AttrR(Int())

    def __init__(self, sub_controller: Controller):
        super().__init__()

        self.attr_on_object = AttrR(Int())

        self.attributes["_attributes_attr"] = AttrR(Int())
        self.attributes["_attributes_attr_equal"] = self.equal_attr

        self.sub_controller = sub_controller


def test_attribute_parsing():
    sub_controller = SomeSubController()
    controller = SomeController(sub_controller)

    assert set(controller.attributes.keys()) == {
        "_attributes_attr",
        "attr_on_object",
        "_attributes_attr_equal",
        "annotated_and_equal_attr",
        "equal_attr",
        "sub_controller",
    }

    assert SomeController.equal_attr is not controller.equal_attr
    assert (
        SomeController.annotated_and_equal_attr
        is not controller.annotated_and_equal_attr
    )

    assert sub_controller.attributes == {
        "sub_attribute": sub_controller.sub_attribute,
    }


def test_conflicting_attributes_and_controllers():
    class ConflictingController(Controller):
        attr = AttrR(Int())

        def __init__(self):
            super().__init__()
            self.sub_controller = Controller()

    controller = ConflictingController()

    with pytest.raises(ValueError, match=r"Cannot add attribute .* existing attribute"):
        controller.attr = AttrR(Float())  # pyright: ignore[reportAttributeAccessIssue]

    with pytest.raises(
        ValueError, match=r"Cannot add sub controller .* existing attribute"
    ):
        controller.attr = Controller()  # pyright: ignore[reportAttributeAccessIssue]

    with pytest.raises(
        ValueError, match=r"Cannot add sub controller .* existing sub controller"
    ):
        controller.sub_controller = Controller()

    with pytest.raises(
        ValueError, match=r"Cannot add attribute .* existing sub controller"
    ):
        controller.sub_controller = AttrR(Int())  # pyright: ignore[reportAttributeAccessIssue]



from pathlib import Path

from fastcs.attributes import AttrR
from fastcs.connections import IPConnection, IPConnectionSettings
from fastcs.controller import Controller
from fastcs.datatypes import String
from fastcs.launch import FastCS
from fastcs.transport.epics.ca.transport import EpicsCATransport
from fastcs.transport.epics.options import EpicsGUIOptions, EpicsIOCOptions


class TemperatureController(Controller):
    device_id = AttrR(String())

    def __init__(self, settings: IPConnectionSettings):
        super().__init__()

        self._ip_settings = settings
        self._connection = IPConnection()

    async def connect(self):
        await self._connection.connect(self._ip_settings)


gui_options = EpicsGUIOptions(
    output_path=Path(".") / "demo.bob", title="Demo Temperature Controller"
)
epics_ca = EpicsCATransport(gui=gui_options, ca_ioc=EpicsIOCOptions(pv_prefix="DEMO"))
connection_settings = IPConnectionSettings("localhost", 25565)
fastcs = FastCS(TemperatureController(connection_settings), [epics_ca])


if __name__ == "__main__":
    fastcs.run()
